#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "SDL.h"

#include "debug.h"

#include "defs.h"

#include "fps.h"

#define PI 3.1415

typedef struct point_t
{
      int x, y;
} point_t;

typedef struct brush_t
{
      int wid, hgt;
      signed char *data;
} brush_t;

typedef struct bob_t
{
      float xangle, yangle;
      float d_xangle, d_yangle;
      // array of points?
      // formula-modification speed?
} bob_t;


typedef struct color_t
{
      int r, g, b;
} color_t;

typedef struct palette_t
{
      int num_colors;
      color_t *colors;
} palette_t;


int wid=320, hgt=240;
int brush_size=0;
int brush_depth=0;
char *brush_filename = "brush.bmp";
int quit = 0;
int frame_delay = 10000;
int num_brushes = NUM_POINTS;
palette_t *global_palette;

int paused=0;


int command_line(int argc, char *argv[]);
SDL_Surface * init_video(int screen_wid, int screen_hgt);
int event_test_loop(SDL_Surface *screen);
int event_test(SDL_Surface *screen);
SDL_Surface * load_brush();

palette_t * make_palette(palette_t *p, int num_colors);

brush_t * make_sine_brush(int wid, int hgt, int depth);
brush_t * make_toroid_brush(int wid, int hgt, int depth);
brush_t * negate_brush(brush_t *brush);
int draw_brush(SDL_Surface *screen, int x, int y, brush_t *brush,
	       unsigned int *height_map, palette_t *palette);
int draw_brush_wrapped(SDL_Surface *screen, int x, int y, brush_t *brush,
	       unsigned int *height_map, palette_t *palette);

int bob_bobs(SDL_Surface *screen, brush_t *brush, palette_t *palette);
int update_screen(SDL_Surface *screen, SDL_Rect *rects, int num_brushes);


//////////////////////////////////////////////////
//
int main(int argc, char *argv[])
{
   SDL_Surface *screen;
   brush_t *brush;
   palette_t *palette;

   debug("main()\n");
   
   srand(time(NULL));
   
   command_line(argc, argv);

   screen = init_video(wid, hgt);

   if(!brush_size)
      brush_size = wid/20;

   if(!brush_depth)
      brush_depth = 8;

   //brush = load_brush();
   brush = make_sine_brush(brush_size, brush_size, brush_depth);
   palette = make_palette(NULL, 1024);
   global_palette = palette;
   

   //color_test(screen);

   FpsStart();
   bob_bobs(screen, brush, palette);
   FpsEnd();

   //event_test_loop(screen);

   // FPS is broken right now...
   //printf("Frames per second: %f\n", Fps());
   ShowFps();
   
   debug("main() done\n");
   return 0;
}

//////////////////////////////////////////////////
//
int command_line(int argc, char *argv[])
{
   int i;

   debug("command_line()\n");
   brush_filename = "brush.bmp";

   for(i=1; i<argc; i++)
   {
      if(0 == strcmp(argv[i], "--help"))
      {
	 printf(helpmessage);
	 exit(0);
      }
      else if(0 == strcmp(argv[i], "--wid"))
      {
	 int tmp;
	 tmp = atoi(argv[++i]);
	 if(tmp > 64  &&  tmp < 0xffff)
	 {
	    wid = tmp;
	 }
      }
      else if(0 == strcmp(argv[i], "--hgt"))
      {
	 int tmp;
	 tmp = atoi(argv[++i]);
	 if(tmp > 64  &&  tmp < 0xffff)
	 {
	    hgt = tmp;
	 }
      }
      else if(0 == strcmp(argv[i], "--delay"))
      {
	 int tmp;
	 tmp = atoi(argv[++i]);
	 if(tmp >= 0)
	 {
	    frame_delay = tmp;
	 }
      }
      else if(0 == strcmp(argv[i], "--brush"))
      {
	 brush_filename = argv[++i];
      }
      else if(0 == strcmp(argv[i], "--num-brushes"))
      {
	 int tmp;
	 tmp = atoi(argv[++i]);
	 if(tmp > 0)
	 {
	    num_brushes = tmp;
	    if(num_brushes > NUM_POINTS)
	    {
	       printf("The maximum number of balls is %i.\n", NUM_POINTS);
	       num_brushes = NUM_POINTS;
	    }
	 }
      }
      else if(0 == strcmp(argv[i], "--brush-size"))
      {
	 int tmp;
	 tmp = atoi(argv[++i]);
	 if(tmp > 0)
	 {
	    brush_size = tmp;
	 }
      }
      else if(0 == strcmp(argv[i], "--brush-depth"))
      {
	 int tmp;
	 tmp = atoi(argv[++i]);
	 if(tmp > 0)
	 {
 	    brush_depth = tmp;
	 }
      }
      else if(0 == strcmp(argv[i], "--foo"))
      {
      }
      else
      {
	 printf("Unrecognized option: %s\n", argv[i]);
      }
   }
   debug("command_line() done\n");
   return 0;
}


//////////////////////////////////////////////////
//
SDL_Surface * init_video(int screen_wid, int screen_hgt)
{
   SDL_Surface *screen = NULL;
   int color_depth=32;

   debug("init_video()\n");
   // Startup SDL
   if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
      fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
      exit(1);
   }
   atexit(SDL_Quit);

   //SDL_EnableUNICODE();
   
   // Set a video mode
   screen = SDL_SetVideoMode(screen_wid, screen_hgt, color_depth,
			     SDL_SWSURFACE);
   if ( screen == NULL ) {
      fprintf(stderr, "Unable to set %ix%ix%i video: %s\n",
	      screen_wid, screen_hgt, color_depth, SDL_GetError());
      exit(1);
   }

   SDL_ShowCursor(0);
   
   debug("init_video() done\n");
   return screen;
}


//////////////////////////////////////////////////
//
int event_test_loop(SDL_Surface *screen)
{
   printf("Event test loop.\n");

   while(1)
   {
      event_test(screen);
      usleep(1);
   }
   
   return 0;
}


//////////////////////////////////////////////////
//
int event_test(SDL_Surface *screen)
{
   SDL_Event event;
   Uint8 *keys;

   debug("event_test()\n");
   while ( SDL_PollEvent(&event) ) {
      switch (event.type) {
	 case SDL_KEYDOWN:
	    printf("The %s key was pressed!\n",
		   SDL_GetKeyName(event.key.keysym.sym));
	    keys = SDL_GetKeyState(NULL);
	    if(keys[SDLK_ESCAPE] == SDL_PRESSED)
	    {
	       quit = 1;
	    }
	    else if(keys[SDLK_SPACE] == SDL_PRESSED)
	    {
	       // Regenerate palette
	       make_palette(global_palette, -1);
	    }
	    else if(keys['p'] == SDL_PRESSED)
	    {
	       if(paused) paused=0;
	       else paused=1;
	    }
	    /*
	    else if(keys[SDLK_KP_PLUS] == SDL_PRESSED)
	    {
	       if(num_balls < MAX_BALLS)
		  num_balls++;
	    }
	    else if(keys[SDLK_KP_MINUS] == SDL_PRESSED)
	    {
	       if(num_balls > 1)
		  num_balls--;
	    }
	    */
	    break;
	 case SDL_MOUSEMOTION:
	    //printf("Mouse moved by %d,%d to (%d,%d)\n", 
	    //	   event.motion.xrel, event.motion.yrel,
	    //	   event.motion.x, event.motion.y);
	    break;
	 case SDL_MOUSEBUTTONDOWN:
	    printf("Mouse button %d pressed at (%d,%d)\n",
		   event.button.button, event.button.x, event.button.y);
	    break;
	 case SDL_QUIT:
	    exit(0);
      }
   }
   
   debug("event_test() done\n");
   return 0;
}

//////////////////////////////////////////////////////////////////////
//
#define PALETTE_SECTIONS 8
palette_t * make_palette(palette_t * p, int num_colors)
{
   palette_t *palette = NULL;
   int i;
   int r, g, b;
   int sample_points = PALETTE_SECTIONS;
   int current_sample = 0;
   int sample_start=0;
   int sample_size;
   color_t sample_data[PALETTE_SECTIONS+1];
   float weight;
   
   debug("make_palette()\n");
   if(p)
   {
      palette = p;
      num_colors = p->num_colors;
   }
   else
   {
      palette = malloc(sizeof(palette_t));
      if(!palette)
      {
	 return NULL;
      }

      palette->colors = malloc(num_colors * sizeof(color_t));
      if(!palette->colors)
      {
	 free(palette);
	 return NULL;
      }

      palette->num_colors = num_colors;
   }
      
   sample_size = num_colors / PALETTE_SECTIONS;

   // Generate sample point data
   memset(sample_data, 0, (PALETTE_SECTIONS+1)*sizeof(color_t));
   for(i=1; i<PALETTE_SECTIONS; i++)
   {
      while(sample_data[i].r + sample_data[i].g + sample_data[i].b < 384)
      {
	 sample_data[i].r = rand()&255;
	 sample_data[i].g = rand()&255;
	 sample_data[i].b = rand()&255;
      }
   }

   // Generate palette based on samples...
   for(i=0; i<num_colors; i++)
   {
      if((i>0)  &&  ((i%sample_size) == 0))
      {
	 current_sample ++;
	 sample_start = i+1;
      }
      //palette->colors[i].r = i/4;
      //palette->colors[i].g = i/2;
      //palette->colors[i].b = i;

      weight = 1.0 - ((float)(i-sample_start) / (float)sample_size);

      r = (weight * sample_data[current_sample].r) +
	 ((1.0-weight) * sample_data[current_sample+1].r);
      g = (weight * sample_data[current_sample].g) +
	 ((1.0-weight) * sample_data[current_sample+1].g);
      b = (weight * sample_data[current_sample].b) +
	 ((1.0-weight) * sample_data[current_sample+1].b);

      palette->colors[i].r = r;
      palette->colors[i].g = g;
      palette->colors[i].b = b;

      /*
      printf("Palette(%3i): %3i %3i %3i, section %i  -> (%3i %3i %3i)\n",
	     i, r, g, b, current_sample,
	     sample_data[current_sample+1].r,
	     sample_data[current_sample+1].g,
	     sample_data[current_sample+1].b
	     );
      */
   }

   debug("make_palette() done\n");
   return palette;
}


//////////////////////////////////////////////////////////////////////
//
// FIXME: wid must equal height
//        Change this, perhaps?
// FIXME: wid and height must be even numbers
brush_t * make_sine_brush(int wid, int hgt, int depth)
{
   brush_t *b = NULL;
   int x, y, z;
   int radius = wid/2;
   float r;
   int square;
   float dist;
   float length;
   int radsquare = radius * radius;

   debug("make_sine_brush()\n");
   if(wid != hgt)
      return NULL;

   // Allocate crap
   b = malloc(sizeof(brush_t));
   if(!b)
   {
      return NULL;
   }

   b->data = malloc(wid*hgt*sizeof(unsigned char));
   if(!b->data)
   {
      free(b);
      return NULL;
   }

   b->wid = wid;
   b->hgt = hgt;

   
   length = (PI/(float)radius/2.0)*(PI/(float)radius/2.0);
   
   for(y=0; y<hgt; y++)
   {
      //printf("Brush: ");
      for(x=0; x<wid; x++)
      {
	 square = (1+x-radius)*(1+x-radius) + (1+y-radius)*(1+y-radius);
	 // If point lies within a circle, use it.
	 if(square < radsquare)
	 {
	    dist = sqrt(square*length);
	    z = ceil(cos(dist) * (float)depth);
	 }
	 // Otherwise, point is outside the circle and should be ignored
	 else
	 {
	    z = 0;
	 }
	 b->data[y*wid + x] = z;
	 //printf("%3i ", z);
      }
      //printf("\n");
   }
   
   debug("make_sine_brush() done\n");
   return b;
}


//////////////////////////////////////////////////////////////////////
//
// FIXME: wid must equal height
//        Change this, perhaps?
// FIXME: wid and height must be even numbers
brush_t * make_toroid_brush(int wid, int hgt, int depth)
{
   brush_t *b = NULL;
   int x, y, z;
   int radius = wid/2;
   float r;
   int square;
   float dist;
   float length, phase;
   int radsquare = radius * radius;

   debug("make_toroid_brush()\n");
   if(wid != hgt)
      return NULL;

   // Allocate crap
   b = malloc(sizeof(brush_t));
   if(!b)
   {
      return NULL;
   }

   b->data = malloc(wid*hgt*sizeof(unsigned char));
   if(!b->data)
   {
      free(b);
      return NULL;
   }

   b->wid = wid;
   b->hgt = hgt;

   
   length = (3.0*PI/(float)radius/4.0)*(3.0*PI/(float)radius/4.0);
   phase = PI + 0.5   ;
   
   for(y=0; y<hgt; y++)
   {
      //printf("Brush: ");
      for(x=0; x<wid; x++)
      {
	 square = (1+x-radius)*(1+x-radius) + (1+y-radius)*(1+y-radius);
	 // If point lies within a circle, use it.
	 if(square < radsquare)
	 {
	    dist = sqrt(square*length);
	    z = ceil(cos(dist+phase) * (float)depth);
	 }
	 // Otherwise, point is outside the circle and should be ignored
	 else
	 {
	    z = 0;
	 }
	 b->data[y*wid + x] = z;
	 //printf("%3i ", z);
      }
      //printf("\n");
   }
   
   debug("make_toroid_brush() done\n");
   return b;
}


//////////////////////////////////////////////////////////////////////
//
brush_t * negate_brush(brush_t *brush)
{
   brush_t *newbrush;
   int i;
   int c;

   debug("negate_brush()\n");
   if(!brush)
      return NULL;

   newbrush = malloc(sizeof(brush_t));
   if(!newbrush)
      return NULL;

   newbrush->data = malloc(brush->wid*brush->hgt);
   if(!newbrush->data)
   {
      free(newbrush);
      return NULL;
   }

   newbrush->wid = brush->wid;
   newbrush->hgt = brush->hgt;
   
   for(i=0; i< (brush->hgt*brush->wid); i++)
   {
      c = brush->data[i];
      newbrush->data[i] = 0-c;
      //printf("Brush: %i\n", newbrush->data[i]);
   }

   debug("negate_brush() done\n");
   return newbrush;
}


//////////////////////////////////////////////////
//
SDL_Surface * load_brush()
{
   SDL_Surface *imagebmp = NULL;
   SDL_Surface *image = NULL;
   int err;

#if 0
   
   /* Load the BMP file into a surface */
   imagebmp = SDL_LoadBMP(ball_filename);
   if ( imagebmp == NULL ) {
      fprintf(stderr, "Couldn't load %s: %s\n", ball_filename, SDL_GetError());
      return NULL;
   }

   if(transparency)
   {
      //printf("Setting transparency...\n");
      err = SDL_SetColorKey(imagebmp, SDL_SRCCOLORKEY, 0);
      if(err < 0)
      {
	 fprintf(stderr, "Couldn't set transparent color.\n");
      }
   }

   /* Convert the image to the video format (maps colors) */
   //printf("Mapping colors...\n");
   image = SDL_DisplayFormat(imagebmp);
   SDL_FreeSurface(imagebmp);
   if ( image == NULL ) {
      fprintf(stderr, "Couldn't map colors: %s\n", SDL_GetError());
      return NULL;
   }

   return image;

#endif
   return NULL;
}

int foo_putpixel(SDL_Surface *screen, int x, int y, int color)
{
   int offset;
   char *poke;
   
   debug("foo_putpixel()\n");
   offset = (int)y*screen->pitch+(int)x*screen->format->BytesPerPixel;
   poke = (unsigned char *)screen->pixels + offset;
   *((unsigned int *)poke) = color;
   SDL_UpdateRect(screen, x, y, 1, 1);

   debug("foo_putpixel() done\n");
   return 0;
   
}


//////////////////////////////////////////////////////////////////////
//
int draw_brush(SDL_Surface *screen, int x, int y, brush_t *brush,
	       unsigned int *height_map, palette_t *palette)
{

   int bx, by;
   // height map, brush
   int offset, boffset;
   unsigned char *screenmem;
   int z;
   unsigned int pixel;
   int wid, hgt;
   int bpp;
   color_t *c;
   
   debug("draw_brush()\n");
   wid = screen->w;
   hgt = screen->h;
   bpp = screen->format->BytesPerPixel;

   
   for(by = 0; by < brush->hgt; by++)
   {
      screenmem = (unsigned char *)screen->pixels
	 + (screen->pitch * (y+by))
	 + (bpp * x);
      offset = ((y+by)*wid) + x;
      boffset = by*brush->wid;
      for(bx = 0; bx < brush->wid; bx++)
      {
	 // determine height
	 //offset = ((y+by)*wid) + x+bx;  // optimized out
	 //boffset = (by*brush->wid) + bx;
	 z = height_map[offset];
	 z += brush->data[boffset];
	 z = z & 0x03ff;
	 height_map[offset] = z;

	 // set color...
	 c = &palette->colors[z];
	 pixel = (((unsigned int)c->r)<<16) +
	    (((unsigned int)c->g)<<8) +
	    (((unsigned int)c->b));

	 // draw pixel...
	 *((unsigned int *)screenmem) = pixel;
	 screenmem += bpp;
	 //((unsigned int *)screen->pixels)[offset] = pixel;

	 offset++;
	 boffset++;
      }
   }
   SDL_UpdateRect(screen, x, y, brush->wid, brush->hgt);
   //SDL_UpdateRect(screen, 0, 0, screen->w, screen->h);
   
   debug("draw_brush() done\n");
   return 0;
}



//////////////////////////////////////////////////////////////////////
//
int draw_brush_wrapped(SDL_Surface *screen, int x, int y, brush_t *brush,
	       unsigned int *height_map, palette_t *palette)
{

   int bx, by;
   int rx, ry;
   // height map, brush
   int offset, boffset;
   unsigned char *screenmem;
   int z;
   unsigned int pixel;
   int wid, hgt;
   int bpp;
   color_t *c;
   
   debug("draw_brush()\n");
   wid = screen->w;
   hgt = screen->h;
   bpp = screen->format->BytesPerPixel;

   while(x < 0)  x += wid;
   while(x >= wid)  x -= wid;
   while(y < 0)  y += hgt;
   while(y >= hgt)  y -= hgt;
   //x %= wid;
   //y %= hgt;
   
   
   for(by = 0; by < brush->hgt; by++)
   {
      ry = y+by;
      if(ry >= hgt) ry -= hgt;
      
      screenmem = (unsigned char *)screen->pixels
	 + (screen->pitch * ry)
	 + (bpp * x);
      offset = (ry*wid) + x;
      boffset = by*brush->wid;
      for(bx = 0; bx < brush->wid; bx++)
      {
	 rx = x+bx;
	 if(rx == wid)
	 {
	    offset -= wid;
	    screenmem -= screen->pitch;
	 }
	 // determine height
	 //offset = ((y+by)*wid) + x+bx;  // optimized out
	 //boffset = (by*brush->wid) + bx;
	 z = height_map[offset];
	 z += brush->data[boffset];
	 z = z & 0x03ff;
	 height_map[offset] = z;

	 // set color...
	 c = &palette->colors[z];
	 pixel = (((unsigned int)c->r)<<16) +
	    (((unsigned int)c->g)<<8) +
	    (((unsigned int)c->b));

	 // draw pixel...
	 *((unsigned int *)screenmem) = pixel;
	 screenmem += bpp;
	 //((unsigned int *)screen->pixels)[offset] = pixel;

	 offset++;
	 boffset++;
      }
   }
   //SDL_UpdateRect(screen, x, y, brush->wid, brush->hgt);
   SDL_UpdateRect(screen, 0, 0, screen->w, screen->h);
   
   debug("draw_brush() done\n");
   return 0;
}



//////////////////////////////////////////////////
//
int bob_bobs(SDL_Surface *screen, brush_t *brush, palette_t *palette)
{
   SDL_Rect *rects;
   SDL_Rect tmp_rect;

   brush_t *erase_brush;
   point_t *points;
   int point_zero=0;
   unsigned int *height_map;
   int i;
   float x, y;
   float maxspeed;
   float deltaspeed;
   float xangle, yangle;
   float d_xangle, d_yangle;
   float xcoef1, xcoef2,
      ycoef1, ycoef2;
   unsigned int pixel;
   unsigned int offset;
   unsigned char *poke;
   int bpp;

   /*
     TODO: Implement variable coefficients somehow!
   float d_xcoef1, d_xcoef2,
      d_ycoef1, d_ycoef2;
   */

   int wid, hgt;

   debug("bob_bobs()\n");
   wid = screen->w - brush->wid;
   hgt = screen->h - brush->hgt;
   //wid = screen->w;
   //hgt = screen->h;
   //wid *= 2;  hgt *= 2;


   erase_brush = negate_brush(brush);
   if(!erase_brush)
   {
      return -1;
   }

   height_map = malloc(sizeof(unsigned int)*screen->w*screen->h);
   if(!height_map)
   {
      free(erase_brush);
      fprintf(stderr, "Error: Ran out of memory.  "
	      "Run again with a smaller resolution.\n");
      return -1;
   }
   memset(height_map, 0, sizeof(unsigned int)*wid*hgt);
   
   
   points = malloc(sizeof(point_t) * NUM_POINTS);
   if(!points)
   {
      free(erase_brush);
      free(height_map);
      fprintf(stderr, "Error: Ran out of memory.  "
	      "Recompile with less points.\n");
      return -1;
   }
   /*
   balls = malloc(sizeof(ball_t) * MAX_BALLS);
   rects = malloc(sizeof(SDL_Rect) * MAX_BALLS * 2);
   if(!balls  ||  !rects)
   {
      fprintf(stderr, "Error: Balls are too big!  "
	      "Recompile with less balls.\n");
      return -1;
   }
   */
   
   
   pixel = SDL_MapRGB(screen->format, 0xff, 0xff, 0xff);
   bpp = screen->format->BytesPerPixel;
   
   /*
   for(i=0; i<MAX_BALLS; i++)
   {
      balls[i].x = rand() % (wid);
      balls[i].y = rand() % (hgt);

      balls[i].xspeed = (rand()%101-50)/50.0;
      balls[i].yspeed = (rand()%101-50)/50.0;

      rects[2*i].w = ball->w;
      rects[2*i].h = ball->h;
      rects[2*i+1].w = ball->w;
      rects[2*i+1].h = ball->h;
   }
   maxspeed = 15.0;
   deltaspeed = 30.0;
   */

   xangle = (float)(rand()%100000) / 8000.0;
   yangle = (float)(rand()%100000) / 8000.0;
   //d_xangle = (float)(rand()%1001 + 500) / (wid*hgt/4);
   //d_yangle = (float)(rand()%1001 + 500) / (wid*hgt/4);
   d_xangle = (float)(rand()%1001 + 500) / 60000.0;
   d_yangle = (float)(rand()%1001 + 500) / 60000.0;
   xcoef1 = (float)(rand()%1001 + 500) / 20000.0;
   xcoef2 = (float)(rand()%1001 + 500) / 20000.0;
   ycoef1 = (float)(rand()%1001 + 500) / 20000.0;
   ycoef2 = (float)(rand()%1001 + 500) / 20000.0;
   
   while(!quit)
   {
      if(paused)
      {
	 event_test(screen);
	 usleep(1);
	 continue;
      }
      
      // Calculate points
      x = (xcoef1 * sin(xangle)) * (xcoef2 * cos(xcoef1 * xangle));
      y = (ycoef1 * sin(yangle)) * (ycoef2 * cos(ycoef1 * yangle));

      x /= (xcoef1 * xcoef2);
      y /= (ycoef1 * ycoef2);
      x = (wid/2) + (wid/2)*(x);
      y = (hgt/2) + (hgt/2)*(y);
      //x = wid + wid*(x);
      //y = hgt + hgt*(y);

      points[point_zero].x = x;
      points[point_zero].y = y;
      
      // Draw points
      draw_brush(screen, x, y, brush, height_map, palette);
      //draw_brush_wrapped(screen, x, y, brush, height_map, palette);
      //foo_putpixel(screen, x, y, 0xffffffff);
      //offset = (int)y*screen->pitch+(int)x*bpp;
      //poke = (unsigned char *)screen->pixels + offset;
      //*((unsigned int *)poke) = pixel;
      //SDL_UpdateRect(screen, x, y, 1, 1);

      if(frames >= (NUM_POINTS-1))
      {
	 int erase_frame;
	 
	 erase_frame = (point_zero + 1 + NUM_POINTS) % NUM_POINTS;

	 draw_brush(screen, points[erase_frame].x, points[erase_frame].y,
		    erase_brush, height_map, palette);
	 //foo_putpixel(screen, points[erase_frame].x, points[erase_frame].y,
	 //	      0xffffffff);
	 //printf("pz: %i   ef: %i\n",
	 //	point_zero, erase_frame);
	 //printf("Erase %i: %i, %i\n", frames,
	 //	points[erase_frame].x, points[erase_frame].y);
      }
      point_zero = (point_zero + 1) % NUM_POINTS;
      
      // change variables
      xangle += d_xangle;
      yangle += d_yangle;
      
      // update screen
      // check input
      event_test(screen);
      // next frame
      frames++;
      if(frame_delay > 0)
      {
	 if((frames&0xf) == 0)
	    usleep(frame_delay);
      }

      if((frames&0xff)  ==  0xff)
      {
	 //memset(screen->pixels, 0, wid*hgt*bpp);
	 //SDL_UpdateRect(screen, 0, 0, wid, hgt);
      }

      /*
      for(i=0; i<num_balls; i++)
      {
	 rects[2*i].x = balls[i].x;
	 rects[2*i].y = balls[i].y;
	 rects[2*i].w = ball->w;
	 rects[2*i].h = ball->h;
	 
	 SDL_FillRect(screen, &rects[2*i], 0);
      }

      for(i=0; i<num_balls; i++)
      {
	 
	 balls[i].x += balls[i].xspeed;
	 balls[i].y += balls[i].yspeed;

	 balls[i].xspeed += (float)((rand()%101-50)/deltaspeed);
	 balls[i].yspeed += (float)((rand()%101-50)/deltaspeed);
	 balls[i].yspeed += maxspeed / 50.0;

	 if(balls[i].xspeed > maxspeed) balls[i].xspeed = maxspeed;
	 if(balls[i].xspeed < -maxspeed) balls[i].xspeed = -maxspeed;
	 if(balls[i].yspeed > maxspeed) balls[i].yspeed = maxspeed;
	 if(balls[i].yspeed < -maxspeed) balls[i].yspeed = -maxspeed;

	 if(balls[i].x >= wid)
	 {
	    balls[i].x = wid;
	    balls[i].xspeed = -balls[i].xspeed;
	 }
	 if(balls[i].y >= hgt)
	 {
	    balls[i].y = hgt;
	    balls[i].yspeed = -balls[i].yspeed;
	 }
	 if(balls[i].x < 0)
	 {
	    balls[i].x = 0;
	    balls[i].xspeed = -balls[i].xspeed;
	 }
	 if(balls[i].y < 0)
	 {
	    balls[i].y = 0;
	    balls[i].yspeed = -balls[i].yspeed;
	 }

	 rects[2*i+1].x = balls[i].x;
	 rects[2*i+1].y = balls[i].y;

	 draw_ball(screen, ball, balls[i].x, balls[i].y);

      }

      update_screen(screen, rects, num_balls * 2);
      event_test(screen);
      frames++;
      if(frame_delay > 0)
	 usleep(frame_delay);
      */
   }

   // Free Willy!
   //free(balls);
   free(points);
   
   debug("bob_bobs() done\n");
   return 0;
}

//////////////////////////////////////////////////
//
int draw_ball(SDL_Surface *screen, SDL_Surface *ball, int x, int y)
{
   SDL_Rect rect;

   rect.x = x;
   rect.y = y;
   rect.w = ball->w;
   rect.h = ball->h;

   SDL_BlitSurface(ball, NULL, screen, &rect);

   //SDL_UpdateRects(screen, 1, &rect);

   return 0;
}

//////////////////////////////////////////////////
//
// TODO: update rectangles instead of the entire screen
int update_screen(SDL_Surface *screen, SDL_Rect *rects, int num_balls)
{
   SDL_UpdateRects(screen, num_balls, rects);

   return 0;
}
